//
//			TCG -- Trader Coalition Game Simulation
//
//	The simulation is made up of four main classes:
//		tcgSim: manages the overall simulation
//		tcgWorld: manages interactions between traders
//		tcgPopulation: manages evolution of traders
//		tcgTraderBot: simulation of a single trader -- either buyer, seller, or both
//
//	In addition, the TraderBot class makes use of an auxillary class, TraderInfo, to
//	store information about other traders, including a history of transactions.
//
//	The program creates a single instance of tcgSim which creates one instance of
//	tcgWorld and one instance of tcgPopulation. The population creates n TraderBot
//	instances.

#include <stdio.h>
#include "simbio.h"

#include "simulat.h"
#include "world.h"
#include "populat.h"
#include "ptype.h"
#include "fsm.h"    
#include "random.h"

class tcgWorld;
class tcgPopulation;
class tcgTradeBot;

class tcgSimulation : public bioSimulation
{ 
public:
	ctor				tcgSimulation (void);
	dtor				~tcgSimulation (void);
	
	virtual void		Init (void);
	virtual void		InitGen (void);
	virtual void		MatchTraders (void);
	virtual void		Trade (void);
	virtual void		UpdateExp (void);
	virtual void		AssessFitness (void);
	virtual void		EvolveGen (void); 
	                                                 
	virtual int			GMax (void) { return _gmax; }
	virtual int			IMax (void) { return _imax; }
	
	virtual int			InitExpPayoff (void) 	{ return _initExpPayoff; } 
	virtual int			RefusalPayoff (void) 	{ return _refusalPayoff; }
	virtual int			WallflowerPayoff (void) { return _wallflowerPayoff; }
	
	virtual int			BuyerQuota (void) 		{ return _buyerQuota; }
	virtual int			SellerQuota (void) 		{ return _sellerQuota; } 
	
	virtual bioList&	TraderList (void) 		{ return _traders; }
	virtual bioList&	BuyerList (void) 		{ return _buyers; }
	virtual bioList&	SellerList (void) 		{ return _sellers; }

	virtual int			DilemmaPayoff (int a, int b);
	virtual void		MediateTrade (tcgTradeBot* seller, tcgTradeBot* buyer);
		
protected:                         
	tcgPopulation*		_thePopulation; 
	
	int					_traderCount;		// total number of tradebots
	int					_sellerCount;		// number of tradebots that are sellers
	int					_buyerCount;		// number of tradebots that are buyers  
	
	int					_sellerQuota;		// maximum offers made
	int					_buyerQuota;		// maximum offers accepted
	
	int					_gmax;				// maximum number of generations
	int					_imax;				// maximum number of trade cycles per generation 
	
	int					_initExpPayoff;		// initial expected payoff
	int					_refusalPayoff;		// rejection payoff     
	int					_wallflowerPayoff;	// wallflower payoff  
                                                                 
	bioList				_traders;			// list of all tradebots                                                                 
	bioList				_buyers;			// list of buyers
	bioList				_sellers;           // list of sellers (a single tradebot may be in all lists)
	
	int					_trades;			// number of trades mediated in one step

											// PD payoffs:
	int					_bothCoop; 			// payoff for both cooperating
	int					_bothDefect;        // payoff for both defecting
	int					_temptation;        // the temptation payoff
	int					_sucker;            // the sucker payoff 
};

class tcgPopulation : public bioPopulation
{
public:
	void				Init (void) {}
	
  	override bioGType*	GenGType (void);
  	override bioPType*	GenPType (bioGType*);
};

class TraderInfo
{
public:
	ctor				TraderInfo (void);
	
	tcgTradeBot* 		trader;        		// who is this information about?
	int					lastPayoff; 		// payoff from last trade
	int					totalPayoff;        // sum of payoff from all trades
	double				expPayoff;          // expected payoff
	long				history;            // history of trader's actions (bitstring)
	int					count;              // number of trades completed with trader
	int					offerStatus;		// 0 = no offer, 1 = offer made, -1 = offer rejected  
	
	enum				{NO_OFFER, OFFER_MADE, OFFER_REJECTED};
};

TraderInfo::TraderInfo (void)
{
	trader 		= 0;
	lastPayoff	= 0;
	totalPayoff	= 0;
	expPayoff 	= 0.0;
	history 	= 0L;
	count 		= 0;  
	offerStatus = NO_OFFER;
}

class tcgTradeBot : public bioPType
{            
public:
	ctor				tcgTradeBot (bioGType* dna);
    dtor				~tcgTradeBot (void);
    
	virtual void		Look (void) {}
	virtual void		InitProgram (void); 
	
	virtual int			IsBuyer (void) { return (_maxOffers > 0); }
	virtual int			IsSeller (void) { return (_maxAccepts > 0); }
	
	virtual void		Init (tcgSimulation*, bioList&, int maxOffers, int maxAccepts);
	virtual void		BecomeBuyer (tcgSimulation*);
	virtual void		BecomeSeller (tcgSimulation*);
	
	virtual void		PrepareOffers (void);
	virtual int			SubmitOffers (void);
	virtual void		AcceptOffers (void);
	virtual void		TakeOffer (tcgTradeBot* bot);
	virtual void		OfferRejected (tcgTradeBot* bot);
	virtual void		PlayPD (void);
	virtual void		Buy (tcgTradeBot* seller, int payoff, int action);
	virtual void		Sell (tcgTradeBot* buyer, int payoff, int action);  
	virtual void		CalcFitness (void);

	virtual int			GetBuyerAction (tcgTradeBot* trader);
	virtual int			GetSellerAction (tcgTradeBot* trader);
	virtual int			GetAction (TraderInfo*);
	virtual int			GetAction (tcgTradeBot* trader);
	
	virtual TraderInfo*	GetTraderInfo (tcgTradeBot* trader);
	virtual TraderInfo*	GetBuyerInfo (tcgTradeBot* trader);
	virtual TraderInfo*	GetSellerInfo (tcgTradeBot* trader);
	virtual TraderInfo*	GetInfo (tcgTradeBot* bot, TraderInfo* list, int count);  
	
	virtual void		Dump (int);
	
protected:                          
	tcgSimulation*		_sim;
	
	int					_maxOffers;			// buyer quota (> 0 only for buyer tradebots)
	int					_maxAccepts;        // seller quota (>0 only for seller tradebots)
	
	int					_score;				// sum of payoffs
	int					_newPay;            // last payoff
	                         
	int					_traderCount;		// size of _traderList 
	TraderInfo*			_traderList;        // information about all other traders
	
	int					_sellerCount;		// size of _sellerList 
	TraderInfo*			_sellerList;        // information about all sellers
	
	int					_buyerCount;		// size of _buyerList
	TraderInfo*			_buyerList;         // information about all buyers
	
	int					_offers;			// number of current offers made
	int					_accepts; 			// number of current offers accepted
	TraderInfo*			_waitList;          // list of offers from buyers
	
	int					_action;			// co-operate or defect?
};  
 

//////////////////////////
//
// tcgSimulation methods
 
tcgSimulation::tcgSimulation (void)
{
	_thePopulation = new tcgPopulation;
}

tcgSimulation::~tcgSimulation (void)
{ 
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
	                                     
	delete _thePopulation;
}

void
tcgSimulation::Init (void)
{                         
	// !!! 	This method will eventually initialize values from an external initialization
	// 		file, but for now we will just hardcode the values.
	         
	_gmax = 2;
	_imax = 10;  
	
	_traderCount 	= 6;
	_sellerCount 	= 4;
	_buyerCount 	= 4;                
	_sellerQuota 	= 1;
	_buyerQuota 	= 1;    
	
	_initExpPayoff		=  1;
	_refusalPayoff		= -1;
	_wallflowerPayoff	=  0;
	           
	// Initialize prisoner dilemma payoffs
	_bothCoop 	=  1;
	_bothDefect = -1;
	_sucker 	= -2;
	_temptation =  2;
		
	_thePopulation->Init(); 
	_thePopulation->SetSize (_traderCount);
}
 
void
tcgSimulation::InitGen (void)
{  
	int						i;
	tcgTradeBot*			bot = 0;
	tcgTradeBot*			seller = 0;
	tcgTradeBot*			buyer = 0;

	ASSERT (_traders.IsEmpty());
	ASSERT (_buyers.IsEmpty());
	ASSERT (_sellers.IsEmpty());
		        
	for (i = 0; i < _traderCount; i++)
	{   
		bot = (tcgTradeBot*) _thePopulation->GetPType(i);
		ASSERT_VALID (bot);
		
		_traders.Append (bot);		
	}                       
		        
	for (i = 0; i < _buyerCount; i++)
	{   
		bot = (tcgTradeBot*) _thePopulation->GetPType(i);
		ASSERT_VALID (bot);
		
		_buyers.Append (bot);		
	}                       
	
	for (i = 0; i < _sellerCount; i++)
	{   
		bot = (tcgTradeBot*) _thePopulation->GetPType(_traderCount-i-1);
		ASSERT_VALID (bot);

		_sellers.Append (bot);		
	}
	 	
	for (_buyers.First(); buyer = (tcgTradeBot*) _buyers.Next(); )
	{
		buyer->BecomeBuyer (this);
	}
	
	for (_sellers.First(); seller = (tcgTradeBot*) _sellers.Next(); )
	{
		seller->BecomeSeller (this);
	}	
}

void
tcgSimulation::MatchTraders (void)
{  
	tcgTradeBot* 		buyer;  
	tcgTradeBot* 		seller;  
	int					offers;
	                                       
	for (_buyers.First(); buyer = (tcgTradeBot*) _buyers.Next(); )
	{
		buyer->PrepareOffers();
	} 
			                                       
	// Tell all buyers to submit offers and all sellers to accept offers
	// until the buyers stop making offers.
	while (1)
	{                    
		offers = 0;
		
		// Tell all buyers to submit their offers...
		for (_buyers.First(); buyer = (tcgTradeBot*) _buyers.Next(); )
		{
			offers += buyer->SubmitOffers();
		} 
		
		if (offers == 0)
		{
			break;
		}
		
		// Tell all buyers to submit their offers...
		for (_sellers.First(); seller = (tcgTradeBot*) _sellers.Next(); )
		{
			seller->AcceptOffers();
		}             
	} 
}

void
tcgSimulation::Trade (void)
{                              
	tcgTradeBot* 		seller;
		
	_trades = 0;
	                                         
	// Tell all sellers to initiate trading (either accept or refuse offers)	                                            	
	for (_sellers.First(); seller = (tcgTradeBot*) _sellers.Next(); )
	{
		seller->PlayPD();
	}
	
	printf ("total trades: %d\n", _trades);
}

void
tcgSimulation::UpdateExp (void)
{    
}

void
tcgSimulation::AssessFitness (void)
{          
	for (int i = 0; i < _traderCount; i++)
	{   
		tcgTradeBot*	bot = (tcgTradeBot*) _thePopulation->GetPType(i);
		ASSERT_VALID (bot);
		
		bot->CalcFitness();		
	}    
}

void
tcgSimulation::EvolveGen (void)
{        
	tcgTradeBot* 		buyer;  
	tcgTradeBot* 		seller; 
		
	for (_buyers.First(); buyer = (tcgTradeBot*) _buyers.Next(); )
	{
		buyer->Dump (1);
	}
	
	for (_sellers.First(); seller = (tcgTradeBot*) _sellers.Next(); )
	{
		seller->Dump (0);
	}	            
	      
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
	                      
	_thePopulation->Breed();    
	printf ("fitness scores... min: %f avg: %f max: %f\n",
				_thePopulation->GetMinFitness(),
				_thePopulation->GetAvgFitness(),
				_thePopulation->GetMaxFitness());
}

int
tcgSimulation::DilemmaPayoff (int mine, int his) 
//: Returns a payoff for the prisoner's dilemma
{       
	if (mine > 0)
	{
		if (his > 0)
		{
			return _bothCoop;
		}
		else
		{            
			return _sucker;
		}
	}
	else              
	{
		if (his > 0)
		{
			return _temptation;
		}
		else
		{            
			return _bothDefect;
		}
	}
}

void
tcgSimulation::MediateTrade (tcgTradeBot* seller, tcgTradeBot* buyer)
//: Mediate a trade between two tradebots.
{ 
	int					sellerAction;
	int					buyerAction;
	                         
	int					sellerPayoff;
	int					buyerPayoff;
		                         
	ASSERT_VALID (seller);
	ASSERT_VALID (buyer);
	
	// Retrieve an action from each tradebot
	sellerAction = seller->GetSellerAction (buyer);
	buyerAction = buyer->GetBuyerAction (seller);
	                                        
	// Determine a payoff for each action	                                        
	sellerPayoff = DilemmaPayoff (sellerAction, buyerAction);
	buyerPayoff = DilemmaPayoff (buyerAction, sellerAction);
                                         
	// Tell each tradebot what happened, allowing them to update their
	// respective information on each other.                                         
	seller->Sell (buyer, sellerPayoff, buyerAction);
	buyer->Buy (seller, buyerPayoff, sellerAction);

	printf ("Trade: (%p, %d, %d) sold to (%p, %d, %d)\n", 
		seller, sellerAction, sellerPayoff,
		buyer, buyerAction, buyerPayoff);
		                                    
	// Increment our trade count.		                                    
	++_trades;		
}

//////////////////////////
//
// tcgPopulation

bioGType*
tcgPopulation::GenGType (void)
//: Generates a 128 byte haploid genotype. 
{
 	bioHapGType*		gtype = new bioHapGType;
	ASSERT_VALID (gtype);
	            
	gtype->SetLength (8 * 128);
	gtype->Randomize();

	return gtype;	
}

bioPType*
tcgPopulation::GenPType (bioGType* dna)
//: Creates a tradebot.
{
	tcgTradeBot*		bot = new tcgTradeBot (dna);
	ASSERT_VALID (bot);	           

	return bot;
}
 
//////////////////////////
//
// tcgTradeBot

tcgTradeBot::tcgTradeBot (bioGType* dna) : bioPType(dna)
//: TradeBot constructor -- initializes member data with default values.
{   
	_sim = 0;
	
	_maxOffers = 0;
	_maxAccepts = 0;
		    
	_score = 0;   
	_offers = 0;
	_accepts = 0;         
	_newPay = 0;
	
	_sellerCount = 0;
	_sellerList = 0;    
	
	_buyerCount = 0;
	_buyerList = 0;    
	_waitList = 0;
	
	_action = 1;                  
	
	InitProgram();
}

tcgTradeBot::~tcgTradeBot (void)
//: Clean up arrays we allocated.
{
	printf ("dtor %p\n", this);
	
	delete [] _sellerList;
	delete [] _buyerList;
	delete [] _waitList;
}

void
tcgTradeBot::InitProgram (void)         
//: Creates a finite state machine (an instance of bioFSM)
//: to use as our program for determing trade actions.
{
	int              len;
    int              array[128];
    bioFSM*	         rule;

    len = _gtype->GetLength()/8;  // get number of bytes in gtype
    ASSERT (len == 128);
    
    // Extract an array of integers from the genotype to initialize
    // the FSM
    for (int i = 0; i < len; i++)
    {
        array[i] = (int) ((bioHapGType*)_gtype)->GetByte(i);
    }
              
	// Create a new instance of FSM with 2 input bits, 5 state bits
	// and one output bit. The input bits are the last two trade
	// actions the other tradebot made with us. Five state bits
	// allows for 32 internal states (more than we need?!). The one
	// output bit is interpreted as 0 = defect, 1 = cooperate.              
    rule = new bioFSM (2, 5, 1);
    ASSERT_VALID (rule);

    rule->Init();
    rule->ReadArray (array, len);

    SetProgram (rule);
}

static int
compTraderInfo (const void* a, const void* b)
//: Compare two TraderInfo instance, sorting on expected payoff.
{
	// sort according to payoff in descending order
	TraderInfo*	p1 = (TraderInfo*) a;
	TraderInfo*	p2 = (TraderInfo*) b;
	                              
	if (p1->expPayoff < p2->expPayoff)
	{
		return 1;
	}
	else if (p1->expPayoff > p2->expPayoff)
	{
		return -1;
	}
	else
	{                                  
		// If the expected payoffs are the same, flip a coin...
		return 0; //(DRand() > 0.5) ? 1 : -1;
	}
}

void
tcgTradeBot::Init (tcgSimulation* sim, bioList& traders, int maxOffers, int maxAccepts)
{
	tcgTradeBot*			tradebot;
	int						i;	     
                                         
	_sim = sim;  
	_maxOffers = maxOffers;
	_maxAccepts = maxAccepts;
	   
	// Allocate an array of information about the other traders in the game.                                         
	_traderCount = traders.Count();
	_traderList = new TraderInfo[_traderCount]; 
	     
	for (i = 0, traders.First(); tradebot = (tcgTradeBot*) traders.Next(); i++)
	{
		_traderList[i].trader = tradebot;
		// If the tradebot is me, the initial payoff is set to -1 so I don't
		// try to trade with myself, otherwise set to initial expected payoff.  
		_traderList[i].expPayoff = (tradebot == this) ? -1 : _sim->InitExpPayoff();
	}
}

void
tcgTradeBot::BecomeBuyer (tcgSimulation* sim)
//: Assume the role of buyer in the game.
{
	tcgTradeBot*			seller;
	int						i;	     
                                         
	_sim = sim;
	 
	bioList&				sellers = _sim->SellerList();
	 
	// Allocate an array of information about the sellers in the game.                                         
	_sellerCount = sellers.Count();
	_sellerList = new TraderInfo[_sellerCount]; 
	     
	for (i = 0, sellers.First(); seller = (tcgTradeBot*) sellers.Next(); i++)
	{
		_sellerList[i].trader = seller;
		// If I'm a seller too, the initial payoff is set to -1 so I don't
		// try to trade with myself, otherwise set to initial expected payoff.  
		_sellerList[i].expPayoff = (seller == this) ? -1 : _sim->InitExpPayoff();
	}
                         
	_maxOffers = _sim->BuyerQuota();  
	ASSERT (_maxOffers <= _sellerCount);   
}

void
tcgTradeBot::BecomeSeller (tcgSimulation* sim)
//: Assume the role of seller in the game.
{                    
	tcgTradeBot*			buyer;
	int						i;	     
	   
	_sim = sim;                             
	
	bioList&				buyers = _sim->BuyerList();
	
	// Allocate an array of information about the buyers in the game
	// and allocate an array to store offers made.                                       
	_buyerCount = buyers.Count();

	_buyerList = new TraderInfo[_buyerCount]; 
	_waitList = new TraderInfo[_buyerCount]; 
	
	for (i = 0, buyers.First(); buyer = (tcgTradeBot*) buyers.Next(); i++)
	{
		_buyerList[i].trader = buyer; 
		// If I'm a buyer too, the initial payoff is set to -1 so I don't
		// try to trade with myself, otherwise set to initial payoff.  
		_buyerList[i].expPayoff = (buyer == this) ? -1 : _sim->InitExpPayoff(); 
	}
                                                         
	_maxAccepts = _sim->SellerQuota();  
	ASSERT (_maxAccepts <= _buyerCount); 
}

void
tcgTradeBot::PrepareOffers (void)
{ 
	for (int i = 0; i < _sellerCount; i++)
	{
		_sellerList[i].offerStatus = TraderInfo::NO_OFFER;
	} 
	
	_offers = 0;
}

int
tcgTradeBot::SubmitOffers (void) 
//: Submit offers to the best sellers. (This message is sent only to buyers.)
{                 
	int				i;  
	int				newOffers = 0;
	 	
	if (_offers == _maxOffers)
	{
		return 0;
	}
	
	// Sort the seller information based on expected payoff.
	qsort (_sellerList, _sellerCount, sizeof (TraderInfo), compTraderInfo);
	
	// Submit offers up to my quota.
	for (i = 0; i < _sellerCount; i++)
	{    
		// If a seller is not tolerable, quit submitting. (All sellers after
		// this one will also be intolerable because the information is sorted.)
		if (_sellerList[i].expPayoff < 0)
		{
			break;
		}

		// If we haven't already been rejected or accepted, make an offer...
		if (_sellerList[i].offerStatus == TraderInfo::NO_OFFER)
		{
			_sellerList[i].trader->TakeOffer (this);
			_sellerList[i].offerStatus = TraderInfo::OFFER_MADE;
			++_offers;
			++newOffers;
		}   

		// If we reached our maximum, quit now.                             
		if (_offers == _maxOffers)
		{
			break;
		}
	} 

	return newOffers;
}

void
tcgTradeBot::TakeOffer (tcgTradeBot* buyer)
//: Take an offer from a buyer and add it to my waiting list. 
//: (This message is sent only to sellers.)
{
	printf ("%p got an offer from %p\n", this, buyer);
	ASSERT (_accepts < _buyerCount);

	TraderInfo*			info = GetBuyerInfo (buyer);
	ASSERT (info != 0);
		
	_waitList[_accepts].trader = buyer;
	_waitList[_accepts].expPayoff = info->expPayoff;
	_accepts++;
}

void
tcgTradeBot::OfferRejected (tcgTradeBot* seller) 
//: Receive a refusal :-(
{
	printf ("%p got a refusal from %p\n", this, seller);

	TraderInfo*		info = GetSellerInfo (seller);
	ASSERT (info != 0);

	--_offers;
	_newPay = _sim->RefusalPayoff(); 
	_score += _newPay;	

	info->lastPayoff = _newPay;
	info->totalPayoff += _newPay;
	info->offerStatus = TraderInfo::OFFER_REJECTED;
}

void
tcgTradeBot::AcceptOffers (void) 
//: Accept offers from the best buyers. (This message will be sent only to sellers.)
{                                                   
	int					i;
	int					max;
	int					rejections = 0;      
	                                                                       
	// Sort the waiting list based on expected payoff.	                                                                        	
	qsort (_waitList, _accepts, sizeof (TraderInfo), compTraderInfo);
	
	// If we got more offers than our quota we will have to refuse some
	// (potentially) tolerable buyers.
	max = (_accepts > _maxAccepts) ? _maxAccepts : _accepts;
	
	for (i = 0; i < max; i++)
	{ 
		// If this guy is not tolerable, quit accepting offers.
		if (_waitList[i].expPayoff < 0) break;    
	}                      
		
	// Refuse the rest.
	for ( ; i < _accepts; i++)
	{ 
		_waitList[i].trader->OfferRejected (this);   
		++rejections;
	}  

	_accepts -= rejections;
}

void
tcgTradeBot::PlayPD (void) 
//: Trade with the best buyers. (This message will be sent only to sellers.)
{                                                   
	ASSERT_VALID (_sim);
	                                                                        
	for (int i = 0; i < _accepts; i++)
	{                          
		// play PD with other 
		_sim->MediateTrade (this, _waitList[i].trader);
	} 
	
	_accepts = 0;
}

int
tcgTradeBot::GetAction (tcgTradeBot* bot)
// Cooperate or defect with this seller?
{
	return GetAction (GetTraderInfo (bot));
}

int
tcgTradeBot::GetAction (TraderInfo* info)
//: Returns a 1 for cooperate, or 0 for defect based on 
//: what our PD rule says.
{               
	const int		memory = 2;
	int				actions[memory];
	
	for (int i = 0; i < memory; i++)
	{
		actions[i] = (int)(info->history >> i) & 0x1;
	}                                   
	
	return _program->Think (actions, memory);
}

int
tcgTradeBot::GetBuyerAction (tcgTradeBot* seller)
// Cooperate or defect with this seller?
{
	return GetAction (GetSellerInfo (seller));
}

int
tcgTradeBot::GetSellerAction (tcgTradeBot* buyer)
// Cooperate or defect with this buyer?
{
	return GetAction (GetBuyerInfo (buyer));
}

void
tcgTradeBot::Buy (tcgTradeBot* seller, int payoff, int action)
//: Update information about a trade.
{
	TraderInfo*		info = GetSellerInfo (seller);
	ASSERT (info != 0);
	                                                 
	_score += payoff; 
	
	// !!! Need a better way to update expected payoff here.
	info->lastPayoff = payoff;
	info->totalPayoff += payoff;
	info->expPayoff = (info->expPayoff + payoff)/2;
	info->history = (info->history << 1) | action; 
	++info->count;
} 

void
tcgTradeBot::Sell (tcgTradeBot* buyer, int payoff, int action)
//: Update information about a trade.
{
	TraderInfo*		info = GetBuyerInfo (buyer);
	ASSERT (info != 0);
	                                                 
	_score += payoff;
	
	// !!! Need a better way to update expected payoff here.
	info->lastPayoff = payoff;
	info->totalPayoff += payoff;
	info->expPayoff = (info->expPayoff + payoff)/2;
	info->history = (info->history << 1) | action;
	++info->count;
} 

void
tcgTradeBot::CalcFitness (void)
{
	_fitness = pow (2.0, _score);
}

TraderInfo*
tcgTradeBot::GetBuyerInfo (tcgTradeBot* bot)
//: Return informatino about specified buyer.
{                          
	return GetInfo (bot, _buyerList, _buyerCount);
}

TraderInfo*
tcgTradeBot::GetSellerInfo (tcgTradeBot* bot)
//: Return informatino about specified seller.
{                          
	return GetInfo (bot, _sellerList, _sellerCount);
}

TraderInfo*
tcgTradeBot::GetInfo (tcgTradeBot* bot, TraderInfo* list, int count)
//: Look up information about the tradebot in the list.
{
	for (int i = 0; i < count; i++)
	{
		if (list[i].trader == bot)
		{
			return &list[i];
		}
	}

	return 0;
}

TraderInfo*
tcgTradeBot::GetTraderInfo (tcgTradeBot* bot)
//: Look up information about the tradebot in the list.
{
	for (int i = 0; i < _traderCount; i++)
	{
		if (_traderList[i].trader == bot)
		{
			return &_traderList[i];
		}
	}

	return 0;
}

void
tcgTradeBot::Dump (int seller)                         
//: Print information about this tradebot.
{
	TraderInfo*			list = seller ? _sellerList : _buyerList;
	int				   	count = seller ? _sellerCount : _buyerCount;

	printf ("\n%p score = %d (%f)\n", this, _score, _fitness);

	for (int i = 0; i < count; i++)
	{
		printf ("%d %p %d %f %d:", i, list[i].trader, 
				list[i].totalPayoff, list[i].expPayoff, list[i].count);
		
		for (int j = 0; j < list[i].count; j++)
		{
			printf ("%s", (list[i].history >> j) & 0x1 ? "C" : "D");
		}                                              
		printf ("\n");
	}
}
 
void
main (void)
{
	tcgSimulation	tcg;  
	int				g, i;
	
	tcg.Init();
	
	for (g = 0; g < tcg.GMax(); g++)
	{
		printf ("\n\nGeneration #%d\n\n", g);
		tcg.InitGen();
		for (i = 0; i < tcg.IMax(); i++)
		{
			printf ("\nTrade cycle #%d\n\n", i);
			tcg.MatchTraders();
			tcg.Trade();
			tcg.UpdateExp();
		}                   
		tcg.AssessFitness();
		tcg.EvolveGen();
	}   
}
